import os

from sastadev.constants import intreebanksfolder
from sastadev.sastatoken import Token, show
from sastadev.smallclauses import bg, getleavestr, smallclauses, word
from sastadev.tokenmd import TokenListMD
from sastadev.treebankfunctions import getnodeyield, getstree
from sastadev.conf import settings


def getfn(dataset, filename):
    result = os.path.join(settings.DATAROOT, dataset, intreebanksfolder, filename)
    return result


testbank = getfn('sctest', "smallclausetest.xml")
schlichtingtreebank = getfn('schlichting', 'TARVB2_ID2.xml')
mieke06 = getfn('miekeplat', "TARSP_MIEKE06_ID.xml")
mieke08 = getfn('miekeplat', "TARSP_MIEKE08_ID.xml")
aurisraw = getfn('Eliska', 'AURIS_ELISKA_ORIGINAL_ID.xml')
tarsp02 = os.path.join(settings.DATAROOT, 'VKLTarsp', intreebanksfolder, 'Tarsp_02.xml')
tarsp06 = os.path.join(settings.DATAROOT, 'VKLTarsp', intreebanksfolder, 'Tarsp_06.xml')
schlichtingall = os.path.join(
    settings.DATAROOT, 'Schlichtingall', intreebanksfolder, 'TREEBANK_SCHLICHTING_CHAT_ID.xml')


def main():
    smalltest = False
    if smalltest:
        fullnames = [testbank]
    else:
        fullnames = [schlichtingtreebank, mieke06, mieke08,
                     aurisraw, tarsp02, tarsp06, schlichtingall]
    for infullname in fullnames:
        print(f'\n{infullname}\n')
        fulltreebank = getstree(infullname)
        if fulltreebank is not None:
            treebank = fulltreebank.getroot()
            for tree in treebank:
                leaves = getnodeyield(tree)
                tokens = [Token(word(leave), bg(leave)) for leave in leaves]
                tokensmd = TokenListMD(tokens, [])
                resultlist = smallclauses(tokensmd, tree)
                if resultlist != []:
                    print('input:  ', getleavestr(leaves))
                    print('output: ', show(resultlist[0].tokens))
                    print('result: ', resultlist[0].metadata)


if __name__ == '__main__':
    main()




from collections import Counter, defaultdict
from io import BytesIO

import xlsxwriter

from sastadev.ASTApostfunctions import (nounlemmaqid, verblemmaqid,
                                        wordcountperutt)
from sastadev.forms import getformfilename

green = '#00FF00'
red = '#FF0000'
orange = '#FFBB9A'
grey = '#B0B0B0'


# green = 'green'
# green = '#006100'
# red = '#9C0006'
# red = 'red'
# orange = '#9C6500'
# orange = 'yellow'


class ExcelForm:
    def __init__(self, tabel, scores):
        self.tabel = tabel
        self.scores = scores


class AstaFormData:
    def __init__(self, noundict, verbdict, vardict, uttlist):
        self.noundict = noundict
        self.verbdict = verbdict
        self.vardict = vardict
        self.uttlist = uttlist


formulatemplate1 = '=IF(B{}="Niet gevuld",B{},(B{}-Tabel!B{})/Tabel!C{})'


def applytemplate1(rowctr):
    result = formulatemplate1.format(rowctr, rowctr, rowctr, rowctr + 1, rowctr + 1)
    return result


formulatemplate2 = '=(B{}-Tabel!B{})/Tabel!C{}'


def applytemplate2(rowctr):
    result = formulatemplate2.format(rowctr, rowctr + 1, rowctr + 1)
    return result


formulatemplate3 = "=IF(ISERROR(ROUND('{}'!{}3,2)),\"Niet gevuld\",ROUND('{}'!{}3,2))"


def applytemplate3(sheet, colchar):
    result = formulatemplate3.format(sheet, colchar, sheet, colchar)
    return result


finietheidsindexscore = "=IF(ISERROR(ROUND(Uitingen!D3,2)),\"Niet gevuld\",ROUND(Uitingen!D3,2))"
pcuscore = "=IF(ISERROR(ROUND(Uitingen!B4/Uitingen!B1,2)),\"Niet gevuld\",ROUND(Uitingen!B4/Uitingen!B1,2))"

tabel = [['', 'GEM', 'SD'],
         ['Aantal zelfstandige naamwoorden', 48, 7.88],
         ['TTR zelfstandige naamwoorden', 0.76, 0.08],
         ['Lexicale werkwoorden', 29, 4.14],
         ['TTR lexical werkwoorden', 0.63, 0.11],
         ['Semantische parafasieën', 0, 0.57],
         ['Fonologische parafasieën', 0, 0.33],
         ['Neologismen', 0, 0],
         ['Aantal koppel/modaalwerkwoorden', 12, 4.15],
         ['MLU', 8.63, 1.74],
         ['Percentage correcte uitingen', 0.93, 0.06],
         ['Finietheidsindex', 0.99, 0.03],
         ['Aantal bijzinnen', 4.8, 2.78]]

scores = [['', 'Score', 'SD'],
          ['Aantal zelfstandige naamwoorden', "='ZNW & WW'!B1", "=(B2-Tabel!B3)/Tabel!C3"],
          ['TTR zelfstandige naamwoorden', applytemplate3('ZNW & WW', 'B'), applytemplate1(3)],
          ['Aantal lexicale werkwoorden', "='ZNW & WW'!F1", applytemplate2(4)],
          ['TTR lexicale werkwoorden', applytemplate3('ZNW & WW', 'F'), applytemplate1(5)],
          ['Semantische parafasieën', '', applytemplate2(6)],
          ['Fonologische parafasieën', '', applytemplate2(7)],
          ['Neologismen', '', '=IF(B8=0,"Normaal","Afwijkend")'],
          ['Aantal koppel/modaalwerkwoorden', '', applytemplate2(9)],
          ['MLU', applytemplate3('Uitingen', 'B'), applytemplate1(10)],
          ['Percentage correcte uitingen', pcuscore, applytemplate1(11)],
          ['Finietheidsindex', finietheidsindexscore, applytemplate1(12)],
          ['Aantal bijzinnen', '=Uitingen!F2', applytemplate2(13)]]

sheet2hiddenrows = [['Totaal', '=SUM(D6:D105)', '', '', '', '=SUM(H6:H105)'],
                    ['Aantal uniek', '=COUNTA(B6:B105)', '', '', '', '=COUNTA(F6:F105)'],
                    ['TTR', '=B2/B1', '', '', '', '=F2/F1'],
                    ['', '', '', '', '', '']
                    ]

sheet3hiddenrows = [['Aantal', '=COUNTA(B6:B105)', '', 'Som', 'Som', 'Som'],
                    ['Totaal', '=SUM(B6:B105)', '', '=SUM(D6:D105)', '=SUM(E6:E105)', '=SUM(F6:F105)'],
                    ['MLU', '=B2/B1', '', '=(D2/(D2+E2))', '', ''],
                    ['Correct', '=COUNTIF(C6:C105,"J")', '', '', '', '']
                    ]

sheet2header = ['Nummer', 'Zelfstandig naamwoord', 'Herhaling', 'Aantal', '', 'Lexicaal werkwoord', 'Herhaling',
                'Aantal']
sheet2colwidths = [10, 25, 20, 10, 5, 25, 20, 10]
sheet3header = ['Uitingsnummer', 'Aantal woorden', 'Correct', "Goede PV's", "Foute en ontbrekende PV's",
                "Aantal bijzinnen", "Bijzonderheden"]


def writetable(tabel, ws, startrow=0, startcol=0, rhformat=None, chformat=None, cellformat=None):
    curcol = startcol
    currow = startrow
    for row in tabel:
        for el in row:
            if currow == startrow:
                theformat = rhformat
            elif curcol == startcol:
                theformat = chformat
            else:
                theformat = cellformat
            ws.write(currow, curcol, el, theformat)
            curcol += 1
        currow += 1
        curcol = startcol


def make_astaform(theform, astadata, target):
    workbook = xlsxwriter.Workbook(target, {"strings_to_numbers": True})
    bold = workbook.add_format({'bold': True})
    dd2 = workbook.add_format({'num_format': '0.00'})
    okformat = workbook.add_format({'bg_color': green})
    wrongformat = workbook.add_format({'bg_color': red})
    unfilledformat = workbook.add_format({'bg_color': orange})
    bggrey = workbook.add_format({'bg_color': grey})

    condformat1a = {'type': 'cell', 'criteria': 'between', 'minimum': -2, 'maximum': 2, 'format': okformat}
    condformat1b = {'type': 'cell', 'criteria': 'not between', 'minimum': -2, 'maximum': 2, 'format': wrongformat}
    condformat1c = {'type': 'cell', 'criteria': 'equal to', 'value': '"Niet gevuld"', 'format': unfilledformat}
    condformat2a = {'type': 'formula', 'criteria': '=$B$8<>0', 'format': wrongformat}
    condformat2b = {'type': 'formula', 'criteria': '=$B$8=0', 'format': okformat}
    condformat2c = {'type': 'cell', 'criteria': 'equal to', 'value': '"Niet gevuld"', 'format': unfilledformat}

    worksheet1 = workbook.add_worksheet('Uitkomstentabel')
    worksheet2 = workbook.add_worksheet('ZNW & WW')
    worksheet3 = workbook.add_worksheet('Uitingen')
    worksheet4 = workbook.add_worksheet('Tabel')

    # worksheet1
    worksheet1.set_column(0, 0, 30)
    worksheet1.set_column(1, 2, 12)
    worksheet1.hide_gridlines(option=2)
    worksheet1.conditional_format('$C$2:$C$7', condformat1a)
    worksheet1.conditional_format('$C$2:$C$7', condformat1b)
    worksheet1.conditional_format('$C$2:$C$7', condformat1c)
    worksheet1.conditional_format('$C$9', condformat1a)
    worksheet1.conditional_format('$C$9', condformat1b)
    worksheet1.conditional_format('$C$9', condformat1c)
    worksheet1.conditional_format('$C$10:$C$13', condformat1a)
    worksheet1.conditional_format('$C$10:$C$13', condformat1b)
    worksheet1.conditional_format('$C$10:$C$13', condformat1c)
    worksheet1.conditional_format('$C$8', condformat2a)
    worksheet1.conditional_format('$C$8', condformat2b)
    worksheet1.conditional_format('$C$8', condformat2c)

    writetable(theform.scores, worksheet1, rhformat=bold, chformat=bold, cellformat=dd2)
    worksheet1.write('$B$6', '', bggrey)
    worksheet1.write('$B$7', '', bggrey)
    worksheet1.write('$B$8', '', bggrey)
    worksheet1.write('$B$9', '', bggrey)

    for (loc, count) in astadata.vardict.items():
        worksheet1.write(loc, count)

    # worksheet2
    writetable(sheet2hiddenrows, worksheet2)

    colctr = 0
    for (el, w) in zip(sheet2header, sheet2colwidths):
        worksheet2.write(4, colctr, el, bold)
        worksheet2.set_column(colctr, colctr, w)
        colctr += 1

    rowctr = 5
    for i in range(100):
        worksheet2.write(rowctr, 0, i + 1)
        rowctr += 1

    rowctr = 5
    wcol = 1
    freqcol = 3
    for (w, cnt) in astadata.noundict.items():
        worksheet2.write(rowctr, wcol, w)
        worksheet2.write(rowctr, freqcol, cnt)
        rowctr += 1

    rowctr = 5
    wcol = 5
    freqcol = 7
    for (w, cnt) in astadata.verbdict.items():
        worksheet2.write(rowctr, wcol, w)
        worksheet2.write(rowctr, freqcol, cnt)
        rowctr += 1

    for rowctr in range(4):
        worksheet2.set_row(rowctr, None, None, {'hidden': True})

    # workasheet3
    writetable(sheet3hiddenrows, worksheet3)

    colctr = 0
    for el in sheet3header:
        worksheet3.write(4, colctr, el, bold)
        colctr += 1
    worksheet3.set_column(0, colctr, 20)

    rowctr = 5
    writetable(astadata.uttlist, worksheet3, startrow=rowctr)

    for rowctr in range(4):
        worksheet3.set_row(rowctr, None, None, {'hidden': True})

    # worksheet4
    worksheet4.set_column(0, 0, 30)

    writetable(theform.tabel, worksheet4, startrow=1, rhformat=bold)

    return workbook, target


def sumctr(ctr):
    result = sum(ctr.values())
    return result


def getvardict(allresults):
    sempar = 'B6'
    phonpar = 'B7'
    neo = 'B8'
    KM = 'B9'
    semparcount = sumctr(allresults.coreresults['A026']) if 'A026' in allresults.coreresults else 0
    phonparcount = sumctr(allresults.coreresults['A008']) if 'A008' in allresults.coreresults else 0
    neocount = sumctr(allresults.coreresults['A022']) if 'A022' in allresults.coreresults else 0
    Kcount = sumctr(allresults.coreresults['A013']) if 'A013' in allresults.coreresults else 0
    Mcount = sumctr(allresults.coreresults['A020']) if 'A020' in allresults.coreresults else 0
    KMcount = Kcount + Mcount
    result = {sempar: semparcount, phonpar: phonparcount, neo: neocount, KM: KMcount}
    return result


def update(resultdict, ctrdict, key, prop):
    if key in ctrdict:
        ctr = ctrdict[key]
        for uttid in ctr:
            resultdict[uttid][prop] = ctr[uttid]


def updatewithctr(resultdict, ctr, prop):
    for uttid in ctr:
        resultdict[uttid][prop] = ctr[uttid]


def intdictget(dct, keyname):
    r1 = dictget(dct, keyname)
    if r1 == '':
        result = 0
    else:
        result = int(r1)
    return result


def dictget(dct, keyname):
    if keyname in dct:
        result = dct[keyname]
    else:
        result = ''
    return result


def resultdict2table(resultdict):
    table = []
    for uttid in resultdict:
        uttid_dict = resultdict[uttid]
        wc = dictget(uttid_dict, 'wc')
        correct = dictget(uttid_dict, 'correct')
        if correct != '':
            correct = 'J'
        allpvs = intdictget(uttid_dict, 'allpvs')
        foutepvs = intdictget(uttid_dict, 'foutepvs')
        okpvs = max(0, allpvs - foutepvs)
        bijzincount = dictget(uttid_dict, 'bijzincount')
        remarks = dictget(uttid_dict, 'remarks')
        paddeduttid = str(uttid).rjust(3, '0')
        newrow = [paddeduttid, wc, correct, okpvs, foutepvs, bijzincount, remarks]
        table.append(newrow)
    sortedtable = sorted(table, key=lambda row: row[0])
    return sortedtable


def getuttlist(allresults):
    resultdict = defaultdict(lambda: defaultdict(int))
    wordcountdict = wordcountperutt(allresults)
    for uttid in wordcountdict:
        resultdict[uttid]['wc'] = wordcountdict[uttid]
    # update(resultdict, allresults.postresults, 'A046', 'wc')
    update(resultdict, allresults.coreresults, 'A004', 'correct')
    update(resultdict, allresults.coreresults, 'A024', 'allpvs')
    subpvs = allresults.coreresults['A032'] if 'A032' in allresults.coreresults else Counter()
    delpvs = allresults.coreresults['A009'] if 'A009' in allresults.coreresults else Counter()
    tijdfoutpvs = allresults.coreresults['A041'] if 'A041' in allresults.coreresults else Counter()
    foutepvs = subpvs + delpvs + tijdfoutpvs
    updatewithctr(resultdict, foutepvs, 'foutepvs')
    update(resultdict, allresults.coreresults, 'A003', 'bijzincount')

    result = resultdict2table(resultdict)
    return result


def astaform(allresults, _, in_memory=False):
    noundict = defaultdict(int)
    verbdict = defaultdict(int)
    allmatches = allresults.allmatches
    # for el in allmatches:
    #     (qid, uttid) = el
    #     if qid == 'A021':
    #         for amatch in allmatches[el]:
    #             # theword = normalizedword(amatch[0])
    #             theword = getattval(amatch[0], 'lemma')
    #             noundict[theword] += 1
    #     if qid == 'A018':
    #         for amatch in allmatches[el]:
    #             # theword = normalizedword(amatch[0])
    #             theword = getattval(amatch[0], 'lemma')
    #             verbdict[theword] += 1
    for (lemma, uttid) in allresults.postresults[nounlemmaqid]:
        noundict[lemma] += 1
    for (lemma, uttid) in allresults.postresults[verblemmaqid]:
        verbdict[lemma] += 1
    vardict = getvardict(allresults)
    uttlist = getuttlist(allresults)
    astadata = AstaFormData(noundict, verbdict, vardict, uttlist)
    theform = ExcelForm(tabel, scores)

    if in_memory:
        target = BytesIO()
    else:
        target = getformfilename(allresults.filename, '_astaformulier')

    theworkbook, target = make_astaform(theform, astadata, target)
    theworkbook.close()
    return target



"""

SASTAdev is a module used for developing SASTA. It takes as input a treebank or a  (corrected) annotation file,
a method (or method filename), and various kinds of reference files.
We originally used the terms *gold* for the reference file, and *platinum* for an improved reference file,
but later changed this to *bronze*  and *silver*, resp.
In the modules and file names the terms gold and platinum have not been replaced everywhere, so both terms  still occur.


All input is supposed to be encoded in UTF8, all output is also generated in UTF8.


Here is the help for this module (python sastadev.py -h)::

    Usage: sastadev.py [options]

    Options:
      -h, --help            show this help message and exit
      -f INFILENAME, --file=INFILENAME
                            Treebank File to be analysed
      -m METHODNAME, --method=METHODNAME
                            Name of the method or (for backwards compatibility)
                            file containing definition of assessment method (SAM)
      -a ANNOTATIONFILENAME, --anno=ANNOTATIONFILENAME
                            SASTA Annotation Format File containing annotations to
                            derive a  reference
      -g GOLDFILENAME, --gold=GOLDFILENAME
                            File containing a gold reference in SASTA Reference
                            Format
      -c GOLDCOUNTSFILENAME, --goldcounts=GOLDCOUNTSFILENAME
                            File containing  gold reference counts in  SASTA
                            Counts Reference Format
      -p PLATINUMINFILENAME, --plat=PLATINUMINFILENAME
                            File containing a platinum reference in SASTA
                            Reference Format
      -i, --impl            Use the implies column of the method
      -l LOGFILENAME, --log=LOGFILENAME
                            File for logging
      --corr=CORR           0=No correction; 1=correction with 1 alternative;
                            n=correction with multiple alternatives (default)
      --mf=METHODFILENAME, --mfile=METHODFILENAME
                            File containing definition of assessment method (SAM)



By using the preferred file and folder structure, the number of parameters to be
specified is minimal.

The preferred file and folder structure is as follows:

 * There is a dataroot where all data reside.  This dataroot can be specified in the configuration module
 * Within the dataroot there is a folder for each dataset (e.g., VKLTARSP for TARSP data provided by VKL)
 * Within each dataset folder the folders are fixed, as follows:

   * Analyses: for storing previous analysis results
   * Analysiscomparison: for reports generated by sastadev on the differences with an earlier analysis
   * Bronze: for bronze reference annotation files
   * Forms: for forms generated by sastadev
   * Indata: for the .cha or .docx input files
   * Intreebanks: for the treebankfiles corresponding to the files in indata
   * Logging: for various logging files generated by sastadev
   * Outtreebanks: for corrected treebanks generated by sastadev
   * Results: for analysis results generated by sastadev
   * Silver: for silver files
   * Silverperm: for the silver annotation store

If the input data are in a standard dataset folder configuration, sastadev can be called by specifying the full path
to the input treebank (using -f) and specifying the method (using -m), e.g.::

    python sastadev.py -f "D:/jodijk/Dropbox/jodijk/Utrecht/Projects/SASTADATA/VKLTarsp/intreebanks/Tarsp_01.xml" -m tarsp

(where 'D:/jodijk/Dropbox/jodijk/Utrecht/Projects/SASTADATA' is the dataroot for user Odijk101)

If the data are not in the standard configuration, one needs to specify one additional parameter to specify the
full path to the bronze reference file. It is assumed that all other input files are in the same folder as the
input treebank, and all output files are generated in that folder.

It generates as output a lot of files. If the input treebank file has the name Tarsp_10.xml, the following files
are generated as output:

* Log file: Tarsp_10.log
* Platinum files:

  * Tarsp_10.platinum-edited.tsv.txt
  * Tarsp_10.platinum.tsv.txt

* Platinum Check files:

  * TSV format: Tarsp_10_platinum.check.tsv.txt
  * XSLX format: Tarsp_10_platinum.check.tsv.xlsx

* Analysis files:

  * TSV format: Tarsp_10_analysis.tsv.txt
  * XLSX format: Tarsp_10_analysis.xlsx

* Grammatical Error related files:

  * Corrected treebank: Tarsp_10_corrected.xml
  * Error Logging: Tarsp_10_errorlogging.xlsx
  * Error Report: Tarsp_10_errorreport.xlsx

* Tarsp_10_countcomparison.tsv.txt
* Form: Tarsp_10_TARSP-Form.xlsx



Sastadev supports the following reference files:

 * for bronze/gold reference files

   * an annotated reference file in Sasta Annotation File (SAF) format (option -a), or
   * a reference file in Sasta Reference File (SRF) format (option -g), or
   * a reference count file in SASTA Reference Count File (RCF) format (option -c)

 * for silver/platinum reference files:

   * a reference file in Sasta  Reference File (SRF) format (option -p)

Sastadev yields as output:

 * an analysis file showing the results and comparing them to the gold and platinum reference if supplied  (in SASTA Development Analysis (SDA) format)
 * a new platinumfile (useful for creating or updating the platinum reference file), in SPR format
 * a platinum check file: list the examples of mismatches between the results and the gold reference
 * a gold counts comparison file (for  each measure the # of the results, the # of the gold reference count, and their difference) (tsv format)

Sastadev logs its actions through:

 * a logfile, and
 * the terminal screen (sys.stderr)

"""

# to do
# -Excel output, cleanup output code

import datetime
import logging
import os
import re
import sys
from collections import Counter, defaultdict
# from altcodes import altcodes
from optparse import OptionParser
from typing import Any, Callable, Dict, List, Pattern, Tuple

import xlsxwriter
from lxml import etree

from sastadev import compounds
from sastadev.allresults import AllResults, scores2counts
from sastadev.conf import settings
from sastadev.constants import (bronzefolder, formsfolder, intreebanksfolder,
                                loggingfolder, outtreebanksfolder,
                                resultsfolder, silverfolder, silverpermfolder)
from sastadev.correcttreebank import (correcttreebank, corrn, errorwbheader,
                                      validcorroptions)
from sastadev.counterfunctions import counter2liststr
from sastadev.external_functions import str2functionmap
from sastadev.goldcountreader import get_goldcounts
from sastadev.macros import expandmacros
from sastadev.methods import Method, defaultfilters
from sastadev.mismatches import exactmismatches, getmarkposition
from sastadev.mksilver import getsilverannotations, permprefix
from sastadev.query import (Query, form_process, is_core, is_pre, is_preorcore,
                            post_process, query_exists, query_inform)
from sastadev.readmethod import itemseppattern, read_method
from sastadev.rpf1 import getevalscores, getscores, sumfreq
from sastadev.SAFreader import (get_golddata, richexact2global,
                                richscores2scores)
from sastadev.sastatypes import (AltCodeDict, ExactResultsDict, FileName,
                                 GoldTuple, MatchesDict, MethodName, QId,
                                 QIdCount, QueryDict, ResultsCounter,
                                 ResultsDict, SynTree, UttId)
from sastadev.SRFreader import read_referencefile
from sastadev.targets import get_mustbedone, get_targets
from sastadev.treebankfunctions import (getattval, getnodeendmap, getuttid,
                                        getxmetatreepositions, getxselseuttid,
                                        getyield, showtree)
from sastadev.xlsx import mkworkbook

listDir = False
if listDir:
    print(dir())
    exit(0)


tarsp = 'tarsp'
stap = 'stap'
asta = 'asta'
gramat = 'gramat'

codepath = os.path.dirname(os.path.abspath(__file__))
methodspath = os.path.join(codepath, 'data', 'methods')

supported_methods = {}
supported_methods[tarsp] = os.path.join(methodspath, 'TARSP Index Current.xlsx')
supported_methods[asta] = os.path.join(methodspath, 'ASTA Index Current.xlsx')
supported_methods[stap] = os.path.join(methodspath, 'STAP_Index_Current.xlsx')

platinumchecksuffix = '_platinum.check.tsv'
platinumcheckeditedsuffix = '_platinum.check-edited.tsv'
platinumsuffix = '.platinum.tsv'
platinumeditedsuffix = '.platinum-edited.tsv'
bronzesuffix = '_bronze'
silversuffix = '_silver'

path2permfolder = silverpermfolder


# target_intarget, target_xsid, target_all = 0, 1, 2
# intargetxpath = '//meta[@name="intarget"]'
# xsidxpath = '//meta[@name="xsid"]'
# intargetvalxpath = './/meta[@name="intarget"]/@value'
# xsidvalxpath = './/meta[@name="xsid"]/@value'

proc = tarsp

logfile = sys.stderr

space = " "
comma = ","
semicolon = ';'
commaspace = ', '
tab = '\t'
pagesep = semicolon
txtext = ".txt"
tsvext = '.tsv'
logext = ".log"
xlsxext = '.xlsx'
samzn = 'samzn'
goldheaderrows = 1
# platinumheaderrows = 1
na = 'na'

getwordsxpath = ".//node[@pt or @pos]"

queryinfoheaderrow = ['id', 'cat', 'subcat', 'item']
queryresultsheaderrow = ['count', 'results', 'GoldCount', 'Goldresults', 'queryExists']
queryRGscoreheaderrow = ['recall', 'precision', 'f1score', 'intersection', 'gold-results', 'results-gold']
queryRPscoreheaderrow = ['platinum', 'plat_recall', 'plat_precision',
                         'plat_f1score', 'plat-results', 'results-plat']
queryGPscoreheaderrow = ['GP recall', 'GP precision', 'GP F1-score', 'GP intersection', 'P minus G', 'G minus P']

resultsheaderrow = queryinfoheaderrow + queryresultsheaderrow + queryRGscoreheaderrow + queryRPscoreheaderrow + queryGPscoreheaderrow
resultsheaderstring = tab.join(resultsheaderrow)
platinumheaderrow = ['id', 'cat', 'subcat', 'item', 'uttids', 'results-gold', 'status', 'remarks']
platinumheaderstring = tab.join(platinumheaderrow)

allmatches: MatchesDict = {}
altcodes: AltCodeDict = {}

emptycounter: Counter = Counter()
invalidqueries: Dict[QId, Exception] = {}


def checkplatinum(goldscores, platinumscores, queries):
    for qid in goldscores:
        if qid in platinumscores:
            # all values of gold must be in platinum
            if query_exists(queries[qid]):
                diff1 = goldscores[qid][2] - platinumscores[qid]
                if diff1 != Counter():
                    settings.LOGGER.warning('{} has goldscores not in platinum: {}'.format(qid, diff1))


def mkerrorreport(errordict, errorreportfilename: str):
    header = ['name', 'count', 'uttid', 'value', 'source', 'cat', 'subcat', 'origutt', 'parsed_as']
    allrows = []
    for item in errordict:
        count = len(errordict[item])
        summaryrow = [item, count]
        allrows.append(summaryrow)
        for instance in errordict[item]:
            instancerow = instance[0:1] + instance[2:]
            fullrow = [item, ''] + instancerow
            allrows.append(fullrow)

    wb = mkworkbook(errorreportfilename, [header], allrows, freeze_panes=(1, 1))
    wb.close()


def erow(cnt: int) -> List[str]:
    result = []
    for i in range(cnt):
        result.append('')
    return result


def getpostval(qid, thepostresults):
    if qid in thepostresults:
        result = thepostresults[qid]
    else:
        result = ''
    return result


# def scores2counts(scores):
#    counts = {}
#    for el in scores:
#        countval = len(scores[el])
#        counts[el] = countval
#    return counts

def sf(number):
    if type(number) == float or type(number) == int:
        result = '{0:.1f}'.format(number)
    else:
        result = number
    return result


def getmarkedutt(m: SynTree, syntree: SynTree) -> str:
    thewordlist = getyield(syntree)
    thepositions = getwordpositions(m, syntree)
    themarkedyield = getmarkedyield(thewordlist, thepositions)
    yieldstr = space.join(themarkedyield)
    return yieldstr


def mark(str: str) -> str:
    result = '*' + str + '*'
    return result


def getwordpositionsold(matchtree: SynTree, syntree: SynTree) -> List[int]:
    positions1 = []
    for node in matchtree.iter():
        if 'pt' in node.attrib:
            if 'end' in node.attrib:
                positions1.append(node.attrib['end'])

    indexednodes = {}
    for node in syntree.iter():
        if 'index' in node.attrib and ('pt' in node.attrib or 'cat' in node.attrib or 'pos' in node.attrib):
            theindex = node.attrib['index']
            indexednodes[theindex] = node

    thequery2 = ".//node[@index and not(@pt) and not(@cat)]"
    try:
        matches2 = matchtree.xpath(thequery2)
    except etree.XPathEvalError as e:
        matches2 = []
    positions2 = []
    for m in matches2:
        positions2 += getwordpositions(m, syntree)
    positions = positions1 + positions2
    result = [int(p) for p in positions]
    return result


def getwordpositions(matchtree: SynTree, syntree: SynTree) -> List[int]:
    # nothing special needs to be done for index nodes since they also have begin and end
    positions = []
    for node in matchtree.iter():
        if 'end' in node.attrib:
            positions.append(node.attrib['end'])
    result = [int(p) for p in positions]
    return result


def getmarkedyield(wordlist: List[str], positions: List[int]) -> List[str]:
    pos = 1
    resultlist = []
    for w in wordlist:
        if pos in positions:
            resultlist.append(mark(w))
        else:
            resultlist.append(w)
        pos += 1
    return resultlist


def update(thedict: Dict[str, GoldTuple], qid: str, goldtuple: GoldTuple):
    (level, item, thecounter) = goldtuple
    if qid in thedict:
        (oldlevel, olditem, oldcounter) = thedict[qid]
        thedict[qid] = (oldlevel, olditem, oldcounter + thecounter)
    else:
        thedict[qid] = goldtuple


def logprint(str: str):
    print(str, file=logfile)
    if logfile != sys.stderr:
        print(str, file=sys.stderr)

##@@ types to be reconsidered


def getitem2levelmap(mapping):
    resultmap: Dict[Any, List[Any]] = {}
    for (item, level) in mapping:
        if item in resultmap:
            resultmap[item].append(level)
        else:
            resultmap[item] = [level]
    return resultmap


def getcompounds(syntree: SynTree) -> List[SynTree]:
    results: List[SynTree] = []
    tlist = syntree.xpath(getwordsxpath)
    for t in tlist:
        w = t.attrib['word']
        if compounds.iscompound(w):
            results.append(t)
    return results


def isxpathquery(query: str) -> bool:
    cleanquery = query.lstrip()
    return cleanquery.startswith('//')


def doqueries(syntree: SynTree, queries: QueryDict, exactresults: ExactResultsDict, allmatches: MatchesDict, criterion: Callable[[Query], bool]):
    global invalidqueries
    uttid = getuttid(syntree)
    # uttid = getuttidorno(syntree)
    omittedwordpositions = getxmetatreepositions(syntree, 'Omitted Word', poslistname='annotatedposlist')
    # print(uttid)
    # core queries
    junk = 0
    for queryid in queries:
        if queryid not in exactresults:
            exactresults[queryid] = []
        thequeryobj = queries[queryid]
        if criterion(thequeryobj):
            if query_exists(thequeryobj):
                thelistedquery = thequeryobj.query
                if isxpathquery(thelistedquery):
                    expandedquery = expandmacros(thelistedquery)
                    thequery = "." + expandedquery
                    try:
                        matches = syntree.xpath(thequery)
                    except etree.XPathEvalError as e:
                        invalidqueries[queryid] = e
                        matches = []
                else:
                    thef = str2functionmap[thelistedquery]
                    matches = thef(syntree)
            else:
                matches = []
                exactresults[queryid] = []
            # matchingids = [uttid for x in matches]
            for m in matches:
                # showtree(m)
                if m is None:
                    showtree(syntree)
                if (queryid, uttid) in allmatches:
                    allmatches[(queryid, uttid)].append((m, syntree))
                else:
                    allmatches[(queryid, uttid)] = [(m, syntree)]
                exactresult = (uttid, int(getattval(m, 'begin')) + 1)
                exactresults[queryid].append(exactresult)
            # if queryid in results:
            #    results[queryid].update(matchingids)
            # else:
            #    results[queryid] = Counter(matchingids)


def docorequeries(syntree: SynTree, queries: QueryDict, results: ExactResultsDict, allmatches: MatchesDict):
    doqueries(syntree, queries, results, allmatches, is_core)


def doprequeries(syntree: SynTree, queries: QueryDict, results: ExactResultsDict, allmatches: MatchesDict):
    doqueries(syntree, queries, results, allmatches, is_pre)


def dopostqueries(allresults: AllResults, postquerylist: List[QId], queries: QueryDict):
    # post queries
    for queryid in postquerylist:
        thequeryobj = queries[queryid]
        if query_exists(thequeryobj):
            thelistedquery = thequeryobj.query

            # it is assumed that these are all python functions
            thef = str2functionmap[thelistedquery]
            result = thef(allresults, queries)
            allresults.postresults[queryid] = result


def codeadapt(c: str) -> str:
    result = c
    result = re.sub(r'\.', r'\\.', result)
    result = re.sub(r'\(', r'\\(', result)
    result = re.sub(r'\)', r'\\)', result)
    result = re.sub(r'\?', r'\\?', result)
    result = re.sub(r'\*', r'\\*', result)
    result = re.sub(r'\+', r'\\+', result)
    result = re.sub(r' ', r'\\s+', result)
    return result


def mkpatterns(allcodes: List[str]) -> Tuple[Pattern, Pattern]:
    basepattern = r''
    sortedallcodes = sorted(allcodes, key=len, reverse=True)
    adaptedcodes = [codeadapt(c) for c in sortedallcodes]
    basepattern = r'' + '|'.join(adaptedcodes) + '|' + itemseppattern
    fullpattern = r'^(' + basepattern + r')*$'
    return (re.compile(basepattern), re.compile(fullpattern))


def get_definedfornonemptygold(goldscores, queries: QueryDict) -> Tuple[int, List[QId]]:
    undefinedqueries = []
    definedfornonemptygoldscore = 0
    for queryid in goldscores:
        if goldscores[queryid] != emptycounter:
            if queryid in queries:
                if queries[queryid].query != '':
                    definedfornonemptygoldscore += 1
                else:
                    undefinedqueries.append(queryid)
    return (definedfornonemptygoldscore, undefinedqueries)


def get_comparison(resultscounts: QIdCount, goldcounts: QIdCount, queries: QueryDict) -> List[Tuple[QId, int, int]]:
    comparison = []
    for qid in queries:
        if qid in goldcounts:
            gold = goldcounts[qid]
            if qid in resultscounts:
                res = resultscounts[qid]
            else:
                res = 0
        else:
            gold = 0
            if qid in resultscounts:
                res = resultscounts[qid]
            else:
                res = 0
        comparison.append((qid, res, gold))
    return comparison


def getmethodfromfile(filename: str) -> str:
    result = ''
    path, base = os.path.split(filename.lower())
    for m in supported_methods:
        if m in base:
            result = m
    if result == '':
        settings.LOGGER.error('No supported method found in filename')
        exit(-1)
    else:
        return result


def treatmethod(methodname: MethodName, methodfilename: FileName) -> Tuple[MethodName, FileName]:
    if methodname is None and methodfilename is None:
        settings.LOGGER.error('Specify a method using -m ')
        exit(-1)
    elif methodname is None and methodfilename is not None:
        resultmethodfilename = methodfilename
        resultmethodname = getmethodfromfile(methodfilename)
        settings.LOGGER.warning('Method derived from the method file name: {}'.format(resultmethodname))
    elif methodname is not None and methodfilename is None:
        if methodname.lower() in supported_methods:
            resultmethodname = methodname.lower()
            resultmethodfilename = supported_methods[methodname]
        else:
            resultmethodfilename = methodname
            resultmethodname = getmethodfromfile(methodname)
            settings.LOGGER.warning('Method derived from the method file name: {}'.format(resultmethodname))
    elif methodname is not None and methodfilename is not None:
        if methodname.lower() in supported_methods:
            resultmethodname = methodname.lower()
            resultmethodfilename = methodfilename
        else:
            settings.LOGGER.error('Unsupported method specified {}'.format(methodname))
            exit(-1)
    return resultmethodname, resultmethodfilename


topnodequery = './/node[@cat="top"]'


def getexactresults(allmatches: MatchesDict) -> ExactResultsDict:
    result: ExactResultsDict = defaultdict(list)
    for (queryid, uttid) in allmatches:
        matchresults = []
        wholeuttmatch = False
        for (m, _) in allmatches[(queryid, uttid)]:
            # @@hier de topnode opzoeken@@
            if m is None:
                position = 0
                settings.LOGGER.error('None match found')
            else:
                topnodes = m.xpath(topnodequery)
                if topnodes != []:
                    positionstr = getattval(topnodes[0], 'begin')
                    wholeuttmatch = True
                else:
                    positionstr = getattval(m, 'begin')
                    wholeuttmatch = getattval(m, 'cat') == 'top'
                try:
                    position = int(positionstr)
                except ValueError:
                    position = 0
                    settings.LOGGER.error('getexactresults ValueError')
                else:
                    if not wholeuttmatch:
                        position += 1
            matchresults.append((uttid, position))
        result[queryid] += matchresults
    return result


def exact2results(exactresults: ExactResultsDict) -> ResultsDict:
    """
    :param exactresults: dictionary key= query id, value is a list of (uttid, position) pairs
    :return: dictionary(key -queryid, value is a counter uttid: count
    """
    results: ResultsDict = {}
    for qid in exactresults:
        uttidlist = [uttid for (uttid, _) in exactresults[qid]]
        resultvalue = Counter(uttidlist)
        results[qid] = resultvalue
    return results


def adaptpositions(rawexactresults: ExactResultsDict, nodeendmap) -> ExactResultsDict:
    newexactresults: ExactResultsDict = {}
    for qid in rawexactresults:
        newlist = []
        for (uttid, position) in rawexactresults[qid]:
            newposition = getmarkposition(position, nodeendmap, uttid)
            newtuple = (uttid, newposition)
            newlist.append(newtuple)
        newexactresults[qid] = newlist
    return newexactresults


def passfilter(rawexactresults: ExactResultsDict, method: Method) -> ExactResultsDict:
    """
    let only those through that satisfy the filter
    :param rawexactresults: dictionary with queryid as key and a Counter as value, exact results
    :param method: Method object
    :return: a filtered version of rawexactresults: results that pass the filter
    """
    exactresults: ExactResultsDict = defaultdict(list)
    queries = method.queries
    for queryid in rawexactresults:
        query = queries[queryid]
        queryfilter = query.special2
        thefilter = method.defaultfilter if queryfilter is None or queryfilter == '' else str2functionmap[queryfilter]
        exactresults[queryid] = [r for r in rawexactresults[queryid] if thefilter(query, rawexactresults, r)]
    return exactresults


#defaulttarsp = r"TARSP Index Current.xlsx"
defaulttarsp = supported_methods[tarsp]


def main():

    parser = OptionParser()
    parser.add_option("-f", "--file", dest="infilename",
                      help="Treebank File to be analysed")
    parser.add_option("-m", "--method", dest="methodname",
                      help="Name of the method or (for backwards compatibility) "
                           "file containing definition of assessment method (SAM)")
    parser.add_option("-a", "--anno", dest="annotationfilename",
                      help="SASTA Annotation Format File containing annotations to derive a  reference")
    parser.add_option("-g", "--gold", dest="goldfilename",
                      help="File containing a gold reference in SASTA Reference Format")
    parser.add_option("-c", "--goldcounts", dest="goldcountsfilename",
                      help="File containing  gold reference counts in  SASTA Counts Reference Format")
    parser.add_option("-p", "--plat", dest="platinuminfilename",
                      help="File containing a platinum reference in SASTA Reference Format")
    parser.add_option("-i", "--impl", dest="includeimplies", action="store_true",
                      help="Use the implies column of the method", default=False)
    parser.add_option("-l", "--log", dest="logfilename",
                      help="File for logging")
    parser.add_option("--corr", dest="corr", default='n',
                      help="0=No correction; 1=correction with 1 alternative; "
                           "n=correction with multiple alternatives (default) ")
    parser.add_option("--mf", "--mfile", dest="methodfilename",
                      help="File containing definition of assessment method (SAM)")

    (options, args) = parser.parse_args()

    if options.corr is None:
        options.corr = corrn
    if options.corr not in validcorroptions:
        validcorrstr = comma.join(validcorroptions)
        settings.LOGGER.error('Illegal value for -c/--corr option: only the following are allowed: {}'.format(validcorrstr))
        exit(1)

    # @ hier ook toestaan dat er een annotatiefile als input komt (.xlsx)-done
    if options.infilename is None:  # an XML file or an.xlsx file
        settings.LOGGER.error('Specify an input treebank file name to analyse (.xml) or the name of an annotationfile (.xlsx)')
        exit(1)
    elif not os.path.exists(options.infilename):
        settings.LOGGER.error('File {} not found. Aborting'.format(options.infilename))
        exit(1)
    (inbase, inext) = os.path.splitext(options.infilename)
    basepath, basefilename = os.path.split(options.infilename)
    corepath, lastfolder = os.path.split(basepath)
    corefilename, inext = os.path.splitext(basefilename)

    if lastfolder == intreebanksfolder:
        intreebankinput = True
        analysespath = os.path.join(corepath, 'analyses')
        bronzepath = os.path.join(corepath, bronzefolder)
        silverpath = os.path.join(corepath, silverfolder)
        outtreebankspath = os.path.join(corepath, outtreebanksfolder)
        resultspath = os.path.join(corepath, resultsfolder)
        silverpermpath = os.path.join(corepath, silverpermfolder)
        loggingpath = os.path.join(corepath, loggingfolder)
        formspath = os.path.join(corepath, formsfolder)

        outpaths = [analysespath, outtreebankspath, resultspath, silverpermpath, loggingpath, formspath, silverpath]

        for outpath in outpaths:
            try:
                os.makedirs(outpath)
            except FileExistsError:
                pass
    else:
        intreebankinput = False
        analysespath = bronzepath = outtreebankspath = resultspath = loggingpath = formspath = basepath
        silverpermpath = os.path.join(basepath, path2permfolder)

    if inext not in ['.xml', '.xlsx']:
        settings.LOGGER.error('Illegal input file type: must be a treebank (.xml) or an annotationfile (.xlsx)')
        exit(1)
    elif inext in ['.xlsx']:
        annotationinput = True
    else:
        annotationinput = False

    if options.logfilename is None:
        options.logfilename = os.path.join(loggingpath, corefilename + logext)

    options.methodname, options.methodfilename = treatmethod(options.methodname, options.methodfilename)

    # Configure logging
    handler = logging.FileHandler(options.logfilename, 'w', encoding='utf8')
    logformat = '%(levelname)s:%(message)s'
    formatter = logging.Formatter(logformat)
    handler.setFormatter(formatter)
    root_logger = settings.LOGGER
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.INFO)
    settings.LOGGER.info('Start of logging')

    if options.annotationfilename is not None and options.goldcountsfilename is not None:
        settings.LOGGER.info('Annotation file and Gold counts file found; gold counts file ignored')
    elif options.goldfilename is not None and options.goldcountsfilename is not None:
        settings.LOGGER.info('Gold Reference file and Gold counts file found; gold counts file ignored')

    # this is not needed anymore because of treatmethod
    #if options.methodfilename is None:  # an xslx file
    #    options.methodfilename = defaulttarsp
    if options.annotationfilename is None:  # an xlsx file
        options.annotationfilename = os.path.join(bronzepath, corefilename + bronzesuffix + xlsxext)
    if options.platinuminfilename is None:
        if intreebankinput:
            options.platinuminfilename = os.path.join(silverpath, corefilename + platinumeditedsuffix + txtext)
        else:
            options.platinuminfilename = inbase + platinumeditedsuffix + txtext

    if options.goldfilename is not None and options.annotationfilename is not None:
        settings.LOGGER.info('annotationfile and goldfile specified. Annotationfile will be used.')

    if options.goldfilename is None:
        if intreebankinput:
            options.goldfilename = os.path.join(bronzepath, corefilename + ".gold" + ".tsv" + ".txt")
        else:
            options.goldfilename = inbase + ".gold" + ".tsv" + ".txt"
    if options.goldcountsfilename is None:
        if intreebankinput:
            options.goldcountsfilename = os.path.join(bronzepath, corefilename + ".goldcounts" + xlsxext)
        else:
            options.goldcountsfilename = inbase + ".goldcounts" + xlsxext

    # @@adapt this so that the method is read in directly as a Method object
    (queries, item2idmap, altcodes, postorformquerylist) = read_method(options.methodfilename)
    defaultfilter = defaultfilters[options.methodname]
    themethod = Method(options.methodname, queries, item2idmap, altcodes, postorformquerylist,
                       options.methodfilename, defaultfilter)

    # print('annotationfilename=', options.annotationfilename, file=sys.stderr )

    # read the annotation reference file if available, otherwise the gold file, otherwise quit
    goldscores = {}
    if options.annotationfilename != '' and os.path.exists(options.annotationfilename):
        allannutts, richexactgoldscores = get_golddata(options.annotationfilename, item2idmap, altcodes, queries,
                                                       options.includeimplies)
        annuttcount = len(allannutts)
        exactgoldscores = richscores2scores(richexactgoldscores)
        richgoldscores = richexact2global(richexactgoldscores)
        goldscores = richgoldscores
        goldcounts = scores2counts(goldscores)
        goldcountfilename = options.annotationfilename + ".goldcount.tsv.txt"
        goldcountfile = open(goldcountfilename, 'w', encoding='utf8')
        for el in goldcounts:
            print(el, goldcounts[el], sep=tab, file=goldcountfile)
        goldcountfile.close()
        reffilename = options.annotationfilename
    elif options.goldfilename != '' and os.path.exists(options.goldfilename):
        goldscores = read_referencefile(options.goldfilename, logfile)
        goldcounts = scores2counts(goldscores)
        reffilename = options.goldfilename
    elif options.goldcountsfilename != '' and os.path.exists(options.goldcountsfilename):
        goldcounts = get_goldcounts(options.goldcountsfilename)
        if goldcounts == {}:
            settings.LOGGER.error('No gold counts found. Aborting')
            exit(-1)
        else:
            reffilename = options.goldcountsfilename
    else:
        settings.LOGGER.error('Neither an annotationfile nor a goldfile, nor a gold count file specified. Aborting')
        exit(1)

    # rawcoreresults = {}
    # exact = True
    rawexactresults: ExactResultsDict = defaultdict(list)

    # @dit aanpassen , voor al de message-done
    if not os.path.exists(options.infilename):
        settings.LOGGER.error('Input treebank or annotationfile {} not found. Aborting'.format(options.infilename))
        exit(1)

    # gather remarks on results of earlier runs, write them to a perm_file  and adapt the silverscore file

    #(pathname, barefilename) = os.path.split(options.infilename)
    #(base, ext) = os.path.splitext(barefilename)
    #(fullbase, _) = os.path.splitext(options.infilename)
    #permpath = os.path.join(pathname, path2permfolder)

    #pathname = basepath
    #barefilename = basefilename
    base = corefilename
    ext = inext
    permpath = silverpermpath
    fullbase = inbase

    try:
        os.makedirs(permpath)
    except FileExistsError:
        pass

    perm_silverfilename = permprefix + corefilename + '.xlsx'
    perm_silverfullname = os.path.join(permpath, perm_silverfilename)
    #
    platinumcheckeditedfullname = os.path.join(resultspath, corefilename + platinumcheckeditedsuffix + '.xlsx')

    platinumoutfilename = os.path.join(resultspath, corefilename + platinumsuffix + txtext)
    platinumcheckfilename = os.path.join(resultspath, corefilename + platinumchecksuffix + txtext)
    silvercheckfilename = os.path.join(resultspath, corefilename + platinumchecksuffix + '.xlsx')

    (platbase, platext) = os.path.splitext(platinumcheckfilename)
    platinumcheckxlfullname = platbase + '.xlsx'

    silverannotationsdict = getsilverannotations(perm_silverfullname, platinumcheckeditedfullname,
                                                 platinumcheckxlfullname, silvercheckfilename,
                                                 platinumoutfilename, options.platinuminfilename, goldscores)

    analysedtrees: List[SynTree] = []
    nodeendmap = {}

    # @vanaf nu gaat het om een treebank, dus hier een if statement toevoegen-done
    if annotationinput:
        allutts, richexactscores = get_golddata(options.infilename, item2idmap, altcodes, queries, options.includeimplies)
        uttcount = len(allutts)
        exactresults = richscores2scores(richexactscores)
    else:
        tree = etree.parse(options.infilename)
        origtreebank = tree.getroot()
        if origtreebank.tag != 'treebank':
            settings.LOGGER.error("Input treebank file does not contain a treebank element")
            exit(-1)
        allutts = {}
        uttcount = 0
        # determine targets
        targets = get_targets(origtreebank)
        treebank, errordict, allorandalts = correcttreebank(origtreebank, targets, options.methodname, options.corr)

        # create the new treebank
        fulltreebank = etree.ElementTree(treebank)
        newtreebankfullname = os.path.join(outtreebankspath, corefilename + '_corrected' + '.xml')
        fulltreebank.write(newtreebankfullname, encoding="UTF8", xml_declaration=False,
                           pretty_print=True)

        # create error file
        errorreportfilename = os.path.join(resultspath, corefilename + '_errorreport' + '.xlsx')
        mkerrorreport(errordict, errorreportfilename)

        # create error logging
        errorloggingfullname = os.path.join(loggingpath, corefilename + '_errorlogging' + '.xlsx')

        allerrorrows: List[str] = []
        for orandalts in allorandalts:
            if orandalts is not None:
                allerrorrows += orandalts.OrigandAlts2rows(corefilename)
        errorwb = mkworkbook(errorloggingfullname, [errorwbheader], allerrorrows, freeze_panes=(1, 1))
        errorwb.close()

        analysedtrees = []
        for syntree in treebank:
            uttcount += 1
            #settings.LOGGER.error('uttcount={}'.format(uttcount))
            mustbedone = get_mustbedone(syntree, targets)
            if mustbedone:
                analysedtrees.append(syntree)
                doprequeries(syntree, queries, rawexactresults, allmatches)
                docorequeries(syntree, queries, rawexactresults, allmatches)

                # uttid = getuttid(syntree)
                uttid = getxselseuttid(syntree)
                # showtree(syntree)
                if uttid in nodeendmap:
                    settings.LOGGER.error('Duplicate uttid in sample: {}'.format(uttid))
                nodeendmap[uttid] = getnodeendmap(syntree)

                # uttno = getuttno(syntree)
                # allutts[uttno] = getyield(syntree)
                allutts[uttid] = getyield(syntree)

        # determine exactresults and apply the filter to catch interdependencies between prequeries and corequeries
        # rawexactresults = getexactresults(allmatches)
        rawexactresults2 = passfilter(rawexactresults, themethod)
        exactresults = adaptpositions(rawexactresults2, nodeendmap)

        #pas hier de allutts en de rawexactresults2 aan om expansies te ontdoen, gebseerd op de nodeendmap
        #@@to be implemented @@ of misschien in de loop hierboven al?

    # @ en vanaf hier kan het weer gemeenschappelijk worden; er met dus ook voor de annotatiefile een exactresults opgeleverd worden
    # @d epostfunctions for lemma's etc moeten mogelijk wel aangepast worden

    # adapt the exactresults  positions to the reference

    coreresults = exact2results(exactresults)

    postresults: Dict[QId, Any] = {}
    allresults = AllResults(uttcount, coreresults, exactresults, postresults, allmatches, options.infilename, analysedtrees,
                            allutts, annotationinput)

    postquerylist: List[QId] = [q for q in postorformquerylist if queries[q].process == post_process]
    formquerylist: List[QId] = [q for q in postorformquerylist if queries[q].process == form_process]

    dopostqueries(allresults, postquerylist, queries)
    dopostqueries(allresults, formquerylist, queries)

    (base, ext) = os.path.splitext(options.infilename)
    outputfullname = os.path.join(resultspath, corefilename + "_analysis" + tsvext + txtext)
    outfile = open(outputfullname, 'w', encoding='utf8')

    outxlsx = os.path.join(resultspath, corefilename + "_analysis" + xlsxext)
    outworkbook = xlsxwriter.Workbook(outxlsx, {"strings_to_numbers": True})
    outworksheet = outworkbook.add_worksheet()
    outstartrow = 0
    outstartcol = 0
    outrowctr = outstartrow
    outworksheet.freeze_panes('E2')

    platinuminfilefound = False
    if os.path.exists(options.platinuminfilename):
        platinuminfilefound = True
        platinumresults = read_referencefile(options.platinuminfilename, logfile)
        checkplatinum(goldscores, platinumresults, queries)
    else:
        settings.LOGGER.info('Platinum file {} not found.'.format(options.platinuminfilename))
        platinumresults = {}

    # platinumoutfilename = base + platinumsuffix + txtext
    platinumoutfile = open(platinumoutfilename, 'w', encoding='utf8')
    # platinumcheckfilename = base + platinumchecksuffix + txtext
    platinumcheckfile = open(platinumcheckfilename, 'w', encoding='utf8')

    countcomparisonfilename = os.path.join(resultspath, corefilename + '_countcomparison' + '.tsv' + '.txt')

    # print the invalid queries
    for q in invalidqueries:
        settings.LOGGER.error("{}: {}: <{}>".format(q, invalidqueries[q], queries[q].query))

    # print the header
    print(resultsheaderstring, file=outfile)
    outworksheet.write_row(outrowctr, outstartcol, resultsheaderrow)
    outrowctr += 1

    # print the platinumheader
    print(platinumheaderstring, file=platinumoutfile)

    # print the results
    qcount = 0
    invalidqcount = 0
    undefinedqcount = 0
    results: Dict[QId, ResultsCounter] = allresults.coreresults
    # exactresults = getexactresults(allmatches)
    exact = True

    pcheaders = [['User1', 'User2', 'User3', 'MoreorLess', 'qid', 'cat', 'subcat', 'item', 'uttid', 'pos', 'utt']]
    allrows = []

    for queryid in results:
        sortedgolduttlist: List[str] = []
        platinumoutresults: Dict[UttId, int] = Counter()
        platinumoutresultsstring = ''
        qcount += 1
        theresults = results[queryid]
        resultstr = counter2liststr(theresults)
        if queryid in goldscores:
            # (goldlevel, golditem, goldcounter) = goldscores[queryid]
            goldcounter = goldscores[queryid][2]
            goldcount = sumfreq(goldcounter)
            sortedgolduttstr = counter2liststr(goldcounter)
        else:
            goldcount = 0
            sortedgolduttstr = ''
        thequery = queries[queryid]
        if query_exists(thequery):
            if queryid not in invalidqueries:
                qex = 'yes'
            else:
                qex = 'invalid'
                invalidqcount += 1
        else:
            qex = 'no'
            undefinedqcount += 1
        if query_exists(thequery) and queryid not in invalidqueries:
            # print(queryid, file=logfile)
            if queryid in goldscores:
                goldcounter = goldscores[queryid][2]
            else:
                goldcounter = Counter()
            (recall, precision, f1score) = getscores(theresults, goldcounter)
            liststargoldstr = counter2liststr(theresults & goldcounter)
            goldminustheresults = goldcounter - theresults
            goldminusliststr = counter2liststr(goldminustheresults)
            theresultsminusgold = theresults - goldcounter
            listminusgoldstr = counter2liststr(theresultsminusgold)
            if platinuminfilefound and queryid in platinumresults:
                theplatinumresults = platinumresults[queryid]
                sortedplatinumliststr = counter2liststr(theplatinumresults)
                liststarplatinumstr = counter2liststr(theresults & theplatinumresults)
                platinumminusliststr = counter2liststr(theplatinumresults - theresults)
                listminusplatinumliststr = counter2liststr(theresults - theplatinumresults)
                (platinumrecall, platinumprecision, platinumf1score) = getscores(theresults, theplatinumresults)

                goldstarplatinumstr = counter2liststr(goldcounter & theplatinumresults)
                platinumminusgoldstr = counter2liststr(theplatinumresults - goldcounter)
                goldminusplatinumstr = counter2liststr(goldcounter - theplatinumresults)
                (gprecall, gpprecision, gpf1score) = getscores(goldcounter, theplatinumresults)
            else:
                sortedplatinumliststr, liststarplatinumliststr, platinumminusliststr, listminusplatinumliststr = '', '', '', ''
                (platinumrecall, platinumprecision, platinumf1score) = (na, na, na)

                goldstarplatinumstr, platinumminusgoldstr, goldminusplatinumstr = '', '', ''
                (gprecall, gpprecision, gpf1score) = (na, na, na)

        else:
            goldcounter = Counter()
            (recall, precision, f1score) = (na, na, na)
            liststargoldstr, goldminusliststr, listminusgoldstr = '', '', ''
            sortedplatinumliststr, liststarplatinumliststr, platinumminusliststr, listminusplatinumliststr = '', '', '', ''
            (platinumrecall, platinumprecision, platinumf1score) = (na, na, na)
            theresultsminusgold = {}
            goldminustheresults = {}
            goldstarplatinumstr, platinumminusgoldstr, goldminusplatinumstr = '', '', ''
            (gprecall, gpprecision, gpf1score) = (na, na, na)

        platinumoutresults = theresults | goldcounter
        platinumoutresultsstring = counter2liststr(platinumoutresults)

        queryinforow = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item]
        queryresultsrow = [str(sumfreq(theresults)), resultstr, str(goldcount), sortedgolduttstr, qex]
        queryRGscorerow = [sf(recall), sf(precision), sf(f1score), liststargoldstr, goldminusliststr, listminusgoldstr]
        queryRPscorerow = [sortedplatinumliststr, sf(platinumrecall), sf(platinumprecision), sf(platinumf1score),
                           platinumminusliststr, listminusplatinumliststr]
        queryGPscorerow = [sf(gprecall), sf(gpprecision), sf(gpf1score), goldstarplatinumstr, platinumminusgoldstr,
                           goldminusplatinumstr]

        fullresultrow = queryinforow + queryresultsrow + queryRGscorerow + queryRPscorerow + queryGPscorerow
        print(tab.join(fullresultrow), file=outfile)
        outworksheet.write_row(outrowctr, outstartcol, fullresultrow)
        outrowctr += 1

        platinumrow = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item,
                       platinumoutresultsstring, listminusgoldstr, '', '']

        print(tab.join(platinumrow), file=platinumoutfile)

        # @with an annotationfile allmatches is empty so we need to redefine newrows (exactmismatches) markedutt (getmarkedutt)-done
        if exact:
            newrows = exactmismatches(queryid, queries, exactresults, exactgoldscores, allmatches, allutts,
                                      platinumcheckfile, silverannotationsdict, annotationinput)
            allrows += newrows
        else:
            if theresultsminusgold != {}:
                print('More examples', file=platinumcheckfile)
            for uttid in theresultsminusgold:
                if (queryid, uttid) in allmatches:
                    for (m, syntree) in allmatches[(queryid, uttid)]:
                        markedutt = getmarkedutt(m, syntree)
                        platinumcheckrow1 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item,
                                             uttid, markedutt]
                        print(tab.join(platinumcheckrow1), file=platinumcheckfile)

            if goldminustheresults != {}:
                print('Missed examples', file=platinumcheckfile)
            for uttid in goldminustheresults:
                if uttid in allutts:
                    uttstr = space.join(allutts[uttid])
                else:
                    settings.LOGGER.warning('uttid {} not in allutts'.format(uttid))
                platinumcheckrow2 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item, uttid,
                                     uttstr]
                print(tab.join(platinumcheckrow2), file=platinumcheckfile)

    #platinumcheckfullname = platinumcheckfile.name
    #(base, ext) = os.path.splitext(platinumcheckfilename)
    # platinumcheckxlfullname = base + '.xlsx'
    wb = mkworkbook(platinumcheckxlfullname, pcheaders, allrows, freeze_panes=(1, 9))
    wb.close()

    # compute the gold postresults
    goldpostresults: Dict[UttId, int] = {}
    goldcounters: Dict[QId, ResultsCounter] = {}
    allgoldmatches: MatchesDict = {}
    for qid in goldscores:
        goldcounters[qid] = goldscores[qid][2]
    allgoldresults = AllResults(uttcount, goldcounters, exactgoldscores, goldpostresults, allgoldmatches, reffilename, [],
                                allannutts, annotationinput)
    dopostqueries(allgoldresults, postquerylist, queries)

    # compute the platinum postresults

    platinumpostresults: Dict[QId, Any] = {}

    # print the postresults
    thepostresults = allresults.postresults
    for queryid in postquerylist:
        resultposval = str(getpostval(queryid, thepostresults))
        goldpostval = str(getpostval(queryid, goldpostresults))
        platinumpostval = str(getpostval(queryid, platinumpostresults))
        if queries[queryid].query != '':
            qex = 'yes'
        else:
            qex = 'no'

        queryinforow = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item]
        queryresultsrow = ['', resultposval, '', goldpostval, qex] + erow(6) + [platinumpostval] + erow(11)

        postrow = queryinforow + queryresultsrow
        postrowstring = tab.join(queryinforow + queryresultsrow)
        print(postrowstring, sep=tab, file=outfile)
        outworksheet.write_row(outrowctr, outstartcol, postrow)
        outrowctr += 1

    # gather overall results, 2 cases: (1)for defined original measure queries only; (2) for all original measure queries

    overallmethods = [(1, 'Overall (defined pre and core queries in the profile)',
                       lambda x: is_preorcore(x) and query_exists(x) and query_inform(x)),
                      (2, 'Overall (all pre and core queries in the profile)',
                       lambda x: is_preorcore(x) and query_inform(x)),
                      (3, 'Overall (original pre and core measures with defined queries only)',
                       lambda x: is_preorcore(x) and query_exists(x)),
                      (4, 'Overall (all original pre and core measures)', lambda x: is_preorcore(x))]

    logheader = ['datetime', 'treebank', 'scorenr,' 'R', 'P', 'F1', 'P-R', 'P-P', 'P-F1', 'GP-R', 'GP-P', 'GP-F1', 'ref',
                 'method']
    logname = 'sastalog.txt'
    logpath = r'.'
    logfullname = os.path.join(logpath, logname)
    biglogfile = open(logfullname, 'a', encoding='utf8')

    exactlynow = datetime.datetime.now()
    now = exactlynow.replace(microsecond=0).isoformat()

    for (ctr, message, queryfunction) in overallmethods:
        # gather resultscount
        resultscount = 0
        for queryid in results:
            thequery = queries[queryid]
            if thequery.original and queryfunction(thequery):
                resultscount += sum(results[queryid].values())

        # gather goldcount
        goldcount = 0
        for queryid in goldscores:
            thequery = queries[queryid]
            goldcounter = goldscores[queryid][2]
            if thequery.original and queryfunction(thequery):
                goldcount += sum(goldcounter.values())

        # gather platinumcount
        platinumcount = 0
        for queryid in platinumresults:
            if queryid in queries:
                thequery = queries[queryid]
                if thequery.original and queryfunction(thequery):
                    platinumcount += sum(platinumresults[queryid].values())
            else:
                settings.LOGGER.warning('Query {} found in platinumresults but not in queries'.format(queryid))

        # resultsgoldintersectiocount
        resultsgoldintersectioncount = 0
        for queryid in results:
            thequery = queries[queryid]
            if thequery.original and queryfunction(thequery):
                if queryid in goldscores:
                    goldcounter = goldscores[queryid][2]
                    intersection = results[queryid] & goldcounter
                    resultsgoldintersectioncount += sum(intersection.values())
                else:
                    pass
                    # settings.LOGGER.warning('Query {} found in results but not in goldscores'.format(queryid))

        # resultsplatinumintersectioncount
        resultsplatinumintersectioncount = 0
        for queryid in results:
            thequery = queries[queryid]
            if thequery.original and queryfunction(thequery):
                if queryid in platinumresults:
                    intersection = results[queryid] & platinumresults[queryid]
                    resultsplatinumintersectioncount += sum(intersection.values())
                else:
                    pass
                    # settings.LOGGER.warning('queryid {} not in platinumresults'.format(queryid))

        # goldplatinumintersectioncount
        goldplatinumintersectioncount = 0
        for queryid in platinumresults:
            if queryid in queries:
                thequery = queries[queryid]
                if thequery.original and queryfunction(thequery):
                    if queryid in goldscores:
                        goldcounter = goldscores[queryid][2]
                        intersection = goldcounter & platinumresults[queryid]
                        goldplatinumintersectioncount += sum(intersection.values())
                    else:
                        pass
                        # settings.LOGGER.warning('Query {} in platinumresults but not in goldscores'.format(queryid))
            else:
                settings.LOGGER.warning('Query {} in platinumresults but not in queries'.format(queryid))

        (recall, precision, f1score) = getevalscores(resultscount, goldcount, resultsgoldintersectioncount)
        (platinumrecall, platinumprecision, platinumf1score) = getevalscores(resultscount, platinumcount,
                                                                             resultsplatinumintersectioncount)
        (gprecall, gpprecision, gpf1score) = getevalscores(goldcount, platinumcount, goldplatinumintersectioncount)

        overallrow = ['', '', '', message, '', '', '', '', '', sf(recall), sf(precision), sf(f1score),
                      '', '', '', '', sf(platinumrecall), sf(platinumprecision), sf(platinumf1score), '', '',
                      sf(gprecall), sf(gpprecision), sf(gpf1score), '', '', '']

        print(tab.join(overallrow), file=outfile)
        outworksheet.write_row(outrowctr, outstartcol, overallrow)
        outrowctr += 1

        logrow = [now, options.infilename, str(ctr), sf(recall), sf(precision), sf(f1score),
                  sf(platinumrecall), sf(platinumprecision), sf(platinumf1score),
                  sf(gprecall), sf(gpprecision), sf(gpf1score),
                  reffilename, options.methodfilename]

        print(tab.join(logrow), file=biglogfile)

    biglogfile.close()
    outfile.close()
    outworkbook.close()
    platinumoutfile.close()
    platinumcheckfile.close()

    resultscounts = scores2counts(results)

    countcomparison = get_comparison(resultscounts, goldcounts, queries)
    if countcomparison != []:
        countcomparisonfile = open(countcomparisonfilename, 'w', encoding='utf8')
        ccheader = ['Measure', 'result', 'gold', 'diff']
        ccheaderstr = tab.join(ccheader)
        print(ccheaderstr, file=countcomparisonfile)
        for (q, r, g) in countcomparison:
            if not (r == 0 and g == 0):
                print(q, r, g, r - g, sep=tab, file=countcomparisonfile)

    definedqcount = qcount - undefinedqcount

    (definedfornonemptygoldscore, undefinedqueries) = get_definedfornonemptygold(goldscores, queries)
    (definedfornonemptygoldcounts, undefinedqueries) = get_definedfornonemptygold(goldcounts, queries)

    lgoldscores = len(goldscores)

    if lgoldscores != 0:
        percentagecompletion1 = definedfornonemptygoldscore / lgoldscores * 100
        percentagecompletion1str = '{0:.1f}%'.format(percentagecompletion1)
    else:
        percentagecompletion1str = 'N/A'

    lgoldcounts = len(goldcounts)
    if lgoldcounts != 0:
        percentagecompletion2 = definedfornonemptygoldcounts / lgoldcounts * 100
        percentagecompletion2str = '{0:.1f}%'.format(percentagecompletion2)
    else:
        percentagecompletion2str = 'N/A'

    finalmessagetemplate1 = '{} measures, {} undefined, {} defined,  of which {} invalid.'
    finalmessagetemplate2 = '{} measures defined for a non empty gold score out of {} ({}).'
    finalmessagetemplate3 = '{} measures defined for a non empty gold count out of {} ({}).'
    print(finalmessagetemplate1.format(qcount, undefinedqcount, definedqcount, invalidqcount))
    print(finalmessagetemplate2.format(definedfornonemptygoldscore, lgoldscores, percentagecompletion1str))
    print(finalmessagetemplate3.format(definedfornonemptygoldcounts, lgoldcounts, percentagecompletion2str))
    print('Undefined queries:', undefinedqueries)
    settings.LOGGER.info("Done!")


if __name__ == '__main__':
    main()



'''
The module *deregularise* provides functions for dealing with overgeneralisations,
wrong overgeneralisations, and certain misspelled overgeneralisations of
verbal inflection.

It provides:

* functions for generating a list of overgeneralised forms, with their corrections and a characterisation of the error made. The main function for this is *makeparadigm*.
Generating these forms and string them in a file is done by the module *deregularise_test*

  .. autofunction:: deregularise::makeparadigm

* functions for finding the correct form for a wrongly inflected verb. The main function for this is *correctinflection*:

  .. autofunction:: deregularise::correctinflection

The module initialises the dictionary *correction* by reading in the file with the name contained in the constant *correctionfilename*

  .. autodata:: deregularise::correctionfilename

  **Remark** The generation of the corrections and the module's testing should be
  separated.

  **Remark** The function does not work perfectly yet for certain past participles (
  *ge* added incorrectly, and for some forms is yields a correct form (past participle
  with e) but not the more plausible past tense singular.

  **Remark** The functions should be extended so that also inflection information is
  returned.

  **Remark** Instead os a single correction a list of corrections should be returned. a selection can be made on the basis of inflectinnal information.

'''
import csv
import os
import re
from typing import Dict, List, Optional, Tuple, cast

from sastadev.conf import settings

tab = '\t'
plussym = '+'

# maybe also add a variant with e- as prefix instead of ge-, and also without ge-


overgen = 'Overgeneralisation'
noge = 'Lacking ge prefix'
nog = 'Prefix ge without onset'
wrongovergen = 'Wrong Overgeneralisation'
wrongen = 'Wrong -en suffix'
nolabel = 'Correct'

chat_errors = {
    'Overgeneralisation': 'm',
    'Lacking ge prefix': 'm',
    'Prefix ge without onset': 'm',
    'Wrong Overgeneralisation': 'm',
    'Wrong -en suffix': 'm'
}

metaarr = {}
metaarr['ge'] = ''
metaarr[''] = noge
metaarr['e'] = nog

#: The constant *correctionfilename* contains the name of the file that contains the
#: corrections. This file is generated by the module deregularise_test.
correctionfilename = 'inflectioncorrection.tsv.txt'

v_prefixes = ['her', 'ver', 'ont', 'be']
separable_prefixes = ['aan', 'aaneen', 'aantoe', 'achter', 'achteraan', 'achterna', 'achterom', 'achterop', 'achterover', 'achteruit',
                      'adem', 'ader', 'af', 'auto', 'bakzeil', 'beet', 'bekend', 'belang', 'bellen', 'betreft',
                      'bezig', 'bij', 'bijeen', 'binnen', 'blijk', 'bloot', 'boek', 'bot', 'boven', 'buiten', 'buitenom', 'daar',
                      'dank', 'deel', 'dicht', 'dienst', 'diep', 'dol', 'dood', 'door', 'droog', 'dubbel', 'dwars', 'eruit',
                      'feest', 'fijn', 'flauw', 'gade', 'garant', 'geheim', 'gelijk', 'geluk', 'gelukkig', 'gereed', 'gering', 'gerust',
                      'gevaar', 'gevangen', 'gewaar', 'geweld', 'glad', 'goed', 'groot', 'hard', 'heen', 'heet', 'heruit', 'hoog', 'in',
                      'ineen', 'ja', 'kaal', 'kaart', 'kapot', 'kennis', 'klaar', 'klein', 'komedie', 'kopje', 'kort', 'krom', 'kuit',
                      'kwaad', 'kwijt', 'lam', 'langs', 'ledig', 'leeg', 'les', 'lief', 'los', 'maat', 'mat', 'mede', 'mee', 'mis', 'model',
                      'mooi', 'na', 'nabij', 'nat', 'neder', 'neer', 'om', 'omhoog', 'omlaag', 'omver', 'onder', 'onderdoor', 'onderuit', 'op',
                      'opeen', 'open', 'over', 'overeen', 'overhoop', 'paard', 'paardje', 'plaats', 'plat', 'post', 'prijs', 'raak', 'recht',
                      'rond', 'samen', 'schaak', 'schadeloos', 'scheef', 'school', 'schoon', 'schoot', 'schuil', 'snor', 'stand', 'steek',
                      'stijf', 'stil', 'stop', 'storm', 'stuk', 'tegemoet', 'tegen', 'tegenover', 'tekeer', 'tekort', 'teleur', 'teloor',
                      'teniet', 'tentoon', 'terecht', 'terneder', 'terneer', 'terug', 'tevreden', 'teweeg', 'tewerk', 'thee', 'thuis', 'tijd', 'toe',
                      'toegang', 'toneel', 'tussen', 'uit', 'uiteen', 'up', 'vaarwel', 'vandaan', 'vanzelf', 'vast', 'veil', 'ver', 'verbeurd',
                      'vet', 'vol', 'voor', 'vooraf', 'voorbij', 'voorop', 'voort', 'vooruit', 'vreemd', 'vrij', 'vuil', 'waar', 'wacht',
                      'warm', 'weder', 'weer', 'weerom', 'weg', 'wel', 'wijs', 'zaak', 'zoek', 'zwart']


sorted_separable_prefixes = sorted(separable_prefixes, key=lambda x: len(x))


def correctinflection(word: str) -> List[Tuple[str, str]]:
    '''
    The function *correctinflection* returns a list of tuples (corrected form,
    metadata) for (wrongly inflected) *word*. It does so by calling the function
    *getcorrections* applied to *word* and the dictionary *correction*.

    .. autodata:: deregularise::correction
       :annotation:

    .. autofunction:: deregularise::getcorrections

    '''

    result = getcorrections(word, correction)
    return result


def map_error(error_type: str) -> str:
    try:
        return chat_errors[error_type]
    except KeyError:
        return error_type


def detect_error(original: str, correction: str) -> Tuple[int, Optional[str]]:
    """Detects an error comparing a text with a correction and returns
    the desired editing distance and the CHAT error code

    Args:
        original (str): transcribed text
        correction (str): correction

    Returns:
        Tuple[int, Optional[str]]: editing distance and error code
    """
    error = None
    for candidate, candidate_error in correctinflection(original):
        if candidate == correction:
            error = map_error(candidate_error)
    if error is not None:
        return 1, cast(str, error)
    else:
        return 0, None


def alt(thestr):
    result = '[' + thestr + ']'
    return result


def plus(thestr):
    result = thestr + '+'
    return result


consonants = 'bcdfghjklmnpqrstvwxz'
vowels = 'aeiouyáéíóúýäëïöüÿàèìòù'
nodiavowels = 'aeiouy'
tremavowels = 'äëïöüÿ'
diphtongs = ['au', 'ei', 'ie', 'ij', 'oe', 'ou', 'ui']
vccvpattern = alt(vowels) + plus(alt(consonants)) + alt(vowels)
vccvre = re.compile(vccvpattern)
tkofschip = ['t', 'k', 'o', 'f', 's', 'ch', 'p', 'sh', 'sj']
irrplussuffix = 'Irregular form plus regular suffix'


def addtrema(vowel):
    ind = nodiavowels.find(vowel)
    if ind >= 0:
        result = tremavowels[ind]
    else:
        result = vowel
    return result


def gendepref(word: str, prefixes: List[str]) -> List[Tuple[str, str]]:
    results = []
    result = None
    for sep in prefixes:
        if word.startswith(sep):
            lsep = len(sep)
            base = word[lsep:]
            result = (sep, base)
            results.append(result)
    result = ('', word)
    results.append(result)
    return results


def desep(word: str, correction: Dict[str, Tuple[str, str]]) \
        -> List[Tuple[str, str, str]]:
    '''
    The function *desep* splits *word* into a triple of 3 strings:

    * a separable prefix if present, otherwise ''
    * a inseparable prefix if present, otherwise ''
    * the remainder, which must occur in the correction dictionary


    Examples:
        * koopte : ('','', 'koopte)
        * verkoopte: ('', 'ver', 'koopte')
        * uitverkoopte: ('uit', 'ver', 'koopte)
        * verdwijnde: ('', '', 'verdwijnde) ('dwijnde' is not in the correction dictionary since 'dwijnen' is not an existing word in Dutch

    Certain prefixes can be both a separable prefix and a nonseparable preix,
    e.g. *voor* as in *voorkomen*. This will currently always be analysed as a separable
    prefix.

    **Remark** we may want to extend it so that it analyses ambiguous cases in multiple ways.

    **Remark** *komen* is not in the list of irregular verbs. It has been added. Its
    also has an irregular stem (*kom* instead of expected *koom*). This requires
    adaptations as well

    A string is considered a separable prefix if it occurs in the constant
    *separable_prefixes*.

    A string is considered a inseparable prefix if it occurs in the constant *v_prefixes*.
    '''
    results = []
    results1 = gendepref(word, sorted_separable_prefixes)
    for (prefix, base) in results1:
        if base in correction:
            results.append((prefix, '', base))
        else:
            results2 = gendepref(base, v_prefixes)
            for (prefix2, base2) in results2:
                if base2 in correction:
                    results.append((prefix, prefix2, base2))
    return results


def finaldegemination(thestr):
    if len(thestr) >= 2 and thestr[-1] == thestr[-2]:
        result = thestr[:-1]
    else:
        result = thestr
    return result


def multiplesyllables(stem):
    match = vccvre.search(stem)
    result = match is not None
    return result


def isConsonant(thechar):
    if len(thechar) != 1:
        result = False
    else:
        result = thechar in consonants
    return result


def isVowel(thechar):
    if len(thechar) != 1:
        result = False
    else:
        result = thechar in vowels
    return result


def isDiphthong(thestr):
    result = len(thestr) == 2 and thestr in diphtongs
    return result


def isVowelGeminate(thestr):
    if len(thestr) != 2:
        result = False
    else:
        result = isDiphthong(thestr) or (isVowel(thestr[1]) and thestr[0] == thestr[1])
    return result


def IJ(thestr):
    result = thestr[-2:] == 'ij'
    return


def ViViC(thestr):
    result = len(thestr) >= 3 and isConsonant(thestr[-1]) and isVowel(thestr[-2]) and thestr[-3] == thestr[-2]
    return result


def VVC(thestr):
    result = len(thestr) >= 3 and isConsonant(thestr[-1]) and isVowelGeminate(thestr[-3:-1])
    return result


def VC1C2(thestr):
    result = len(thestr) >= 3 and isConsonant(thestr[-1]) and isConsonant(thestr[-2]) and thestr[-1] != thestr[-2]
    return result


def VC1C1(thestr):
    result = len(thestr) >= 3 and isConsonant(thestr[-1]) and isConsonant(thestr[-2]) and thestr[-1] == thestr[-2]
    return result


def CVC(thestr):
    result1 = len(thestr) >= 3 and isConsonant(thestr[-1]) and isVowel(thestr[-2]) and isConsonant(thestr[-3])
    result2 = len(thestr) == 2 and isConsonant(thestr[-1]) and isVowel(thestr[-2])  # eten
    result = result1 or result2
    return result


def CV(thestr):
    result = len(thestr) >= 2 and isConsonant(thestr[-2]) and isVowel(thestr[-1])
    return result


def dup(thestr):
    return thestr + thestr


def endsin(stem, thechar):
    return stem[-1] == thechar


def startswithprefix(stem):
    result1 = len(stem) >= 2 and stem[0:2] in ['be', 'ge'] and multiplesyllables(stem)
    result2 = len(stem) >= 3 and stem[0:3] in v_prefixes and multiplesyllables(stem)
    result = result1 or result2
    return result


def makepastsg(stem, stemFS):
    if stem[-1] in tkofschip or stem[-2:-1] in tkofschip:
        result = stemFS + 'te'
    else:
        result = stemFS + 'de'

    return result, overgen


def makepastpl(stem, stemFS):
    form, meta = makepastsg(stem, stemFS)
    result = form + 'n'
    return result, meta


def makepastpastsg(past, stem):
    if stem[-1] in tkofschip or stem[-2:-1] in tkofschip:
        result = past + 'te'
    else:
        result = past + 'de'
    return result, irrplussuffix


def makepastpastpl(past, stem):
    form, meta = makepastpastsg(past, stem)
    result = form + 'n'
    return result, meta


def makepastpart(stem, stemFS, takesge, prefix='ge'):
    metalabel = overgen
    if stem[-1] in tkofschip or stem[-2:-1] in tkofschip:
        result = stemFS + 't'
    elif CV(stem):
        result = stem + stem[-1] + 'd'
    else:
        result = stemFS + 'd'
    if takesge:
        result = prefix + result
    # no verbs starting with vowel that we must take into account

    if takesge:
        metalabel = plussym.join([metalabel, metaarr[prefix]])

    result = finaldegemination(result)
    return result, metalabel


def makewrongenpastpart(stem, stemFS, takesge, prefix='ge'):
    if ViViC(stemFS):  # maak -> gemaken
        result = stemFS[:-2] + stemFS[-1] + 'en'
    elif CVC(stemFS):   # lik -> gelikken
        result = stemFS + stemFS[-1:] + 'en'
    elif len(stemFS) >= 2 and isDiphthong(stemFS[-2:]):    # doe -> gedoen, zij -> gezijn, but not ei eu?  @@!
        result = stemFS + 'n'
    elif len(stemFS) >= 2 and isConsonant(stemFS[-2]) and isVowel(stemFS[-1]):  # ga -> gegaan
        result = stemFS + stemFS[-1] + 'n'
    else:  # kijk -> gekijken
        result = stemFS + 'en'
    if takesge:
        if prefix == '':
            adaptedresult = result
        elif result[0] in vowels:  # verbs starting with vowel that we must take into account; Yes eten: geëet!
            adaptedresult = addtrema(result[0]) + result[1:]
        else:
            adaptedresult = result
        result = prefix + adaptedresult
    return result, wrongen


def makewrongpastpart(stem, stemFS, takesge, prefix='ge'):
    metalabel = wrongovergen
    if stem[-1] in tkofschip or stem[-2:] in tkofschip:
        result = stemFS + 't'
    elif CV(stem):
        result = stem + stem[-1] + 't'
    elif stem[-1] not in 'd':
        result = stemFS + 't'
    else:
        result = stemFS + 'd'
    if takesge:
        if prefix == '':
            adaptedresult = result
        elif result[0] in vowels:  # verbs starting with vowel that we must take into account; Yes eten: geëet!
            adaptedresult = addtrema(result[0]) + result[1:]
        else:
            adaptedresult = result
        result = prefix + adaptedresult
    result = finaldegemination(result)
    if takesge:
        metalabel = plussym.join([metalabel, metaarr[prefix]])
    return result, overgen


def getcorrections(thestr: str, correction: Dict[str, Tuple[str, str]]) -> List[Tuple[str, str]]:
    '''

    The function *getcorrections* returns a list of tuples (correct form, metadata) for a (wrongly) inflected word *thestr* using the dictionary *correction*.

    * If *thestr* is contained in the *correction* ( e.g. *loopte*), it appends the associated value (*liep*, *Overgeneralisation*) to the resultlist.
    * Otherwise, it splits the word up into a separable prefix, a nonseparable prefix, and a base by means of the function *desep*.
    * If the function *desep* returns an empty list, *getcorrections* also returns an empty list.

      .. autofunction:: deregularise::desep

      **Remark** The prefix *ge* is incorrectly added to verbs with inseparable
      prefixes, e.g. *vervald* is incorrectly mapped on *vergevallen* (instead of on
      *vervallen*).

    '''
    results = []
    if thestr in correction:
        result = correction[thestr]
        results.append(result)
    else:
        sepresults = desep(thestr, correction)
        for (sep, pref, base) in sepresults:
            if pref + base in correction:
                basestr, meta = correction[pref + base]
                result = sep + basestr, meta  # voor bijv. uitgeloopt
                results.append(result)
            elif base in correction:    # voor bijv uitverloofd,
                basestr, meta = correction[base]
                result = sep + pref + basestr, meta
                results.append(result)
            else:
                geroot = 'ge' + base
                if geroot in correction:
                    basestr, meta = correction[geroot[2:]]
                    result = sep + pref + basestr, meta
                    results.append(result)
                else:
                    result = thestr, nolabel
                    results.append(result)
    results = [(wstr.strip(), meta) for wstr, meta in results]
    return results


def getstems(el):
    base = el[:-2]
    if el[-3:] == 'aan':  # gaan, slaan , staan
        stem = el[:-2]
    elif el[-3:] == 'oen':  # doen
        stem = el[:-1]
    elif el[-3:] == 'ijn':  # zijn
        stem = el[:-1]
    elif IJ(base):  # vrijen
        stem = base
    elif VVC(base):       # gieten
        stem = base
    elif VC1C2(base):   # drinken
        stem = base
    elif VC1C1(base):  # bakken
        stem = el[:-3]
    elif CVC(base):     # dragen
        stem = base[:-2] + dup(base[-2:-1]) + base[-1:]
    else:
        stem = 'niet voorzien: ' + el
    if endsin(stem, 'v'):
        stemFS = stem[:-1] + 'f'
    elif endsin(stem, 'z'):
        stemFS = stem[:-1] + 's'
    else:
        stemFS = stem

    if startswithprefix(stem):
        takesge = False
    else:
        takesge = True
    return stem, stemFS, takesge


def makepastpartwithe(stem, stemFS, takesge, prefix='ge'):
    pastpart, metalabel = makepastpart(stem, stemFS, takesge, prefix)
    if pastpart[-2:] == 'en':
        result = pastpart
    elif ViViC(pastpart):   # geloot -> gelote
        result = pastpart[:-3] + pastpart[-2:] + 'e'
    elif CVC(pastpart):  # gelet -> gelette
        result = pastpart + pastpart[-1] + 'e'
    else:
        result = pastpart + 'e'
    return result, metalabel


def makeparadigm(word, forms):
    '''
    The function *makeparadigm* makes a full verbal paradigm for *word* if it occurs in
    *forms* as if *word* (e.g. *vallen*) has a regular inflection (e.g. *valde*,
    *gevald*),  as well
    as forms of incorrect overgeneralisation (e.g. *vielde*) including some
    inflected words with spelling errors (e.g. *gevalt* instead of *gevald*) and the
    prefix *e* instead of *ge* for past participles (e.g. *evald*) . It returns a list of
    triples for each of these inflected words of the form (wrong, meta, good):

    * *wrong* is the incorrect form (e.g. *valde*)
    * *meta* is metadata about the incorrect form (e.g. 'Overgeneralisation')
    * *good* is the correct form (e.g. *viel*)

    This function is not used by SASTA at runtime, but separately in the module
    *deregularise_test* to generate the  correction dictionary that is used by SASTA at runtime.

    '''

    goodpastsg = forms[word][1]
    goodpastpl = forms[word][2]
    goodpastpart = forms[word][3]
    goodpastpartwithe = goodpastpart + 'e'

    (stem, stemFS, takesge) = getstems(word)

    triples = []

    regularpastsg, metalabel = makepastsg(stem, stemFS)
    triples.append((regularpastsg, metalabel, goodpastsg))

    regularpastpl, metalabel = makepastpl(stem, stemFS)
    triples.append((regularpastpl, metalabel, goodpastpl))

    # sliepte but exclude bande from bandde  wat about hieldden?
    if goodpastsg[-2:] not in {'de', 'te'}:
        wrongpastsg, metalabel = makepastpastsg(goodpastsg, stem)
        triples.append((wrongpastsg, metalabel, goodpastsg))

    # sliepten
    if goodpastsg[-2:] not in {'de', 'te'}:
        wrongpastpl, metalabel = makepastpastpl(goodpastsg, stem)
        triples.append((wrongpastpl, metalabel, goodpastpl))

    # perfect participles
    pastparticiple, metalabel = makepastpart(stem, stemFS, takesge)
    triples.append((pastparticiple, metalabel, goodpastpart))

    epastparticiple, metalabel = makepastpart(stem, stemFS, takesge, prefix='e')
    triples.append((epastparticiple, metalabel, goodpastpart))

    zeropastparticiple, metalabel = makepastpart(stem, stemFS, takesge, prefix='')
    triples.append((zeropastparticiple, metalabel, goodpastpart))

    # perfect participle with e
    pastpartwithe, metalabel = makepastpartwithe(stem, stemFS, takesge)
    triples.append((pastpartwithe, metalabel, goodpastpartwithe))

    epastpartwithe, metalabel = makepastpartwithe(stem, stemFS, takesge, prefix='e')
    triples.append((epastpartwithe, metalabel, goodpastpartwithe))

    # put off temporarily because past tense are more important
    # zeropastpartwithe, metalabel = makepastpartwithe(stem, stemFS, takesge, prefix='')
    # triples.append((zeropastpartwithe, metalabel, goodpastpartwithe))

    # perfect participle misspelled t ipv d gevalt
    wrongpastpart, metalabel = makewrongpastpart(stem, stemFS, takesge)
    triples.append((wrongpastpart, metalabel, goodpastpart))

    ewrongpastpart, metalabel = makewrongpastpart(stem, stemFS, takesge, prefix='e')
    triples.append((ewrongpastpart, metalabel, goodpastpart))

    zerowrongpastpart, metalabel = makewrongpastpart(stem, stemFS, takesge, prefix='')
    triples.append((zerowrongpastpart, metalabel, goodpastpart))

    wrongenpastpart, metalabel = makewrongenpastpart(stem, stemFS, takesge)
    triples.append((wrongenpastpart, metalabel, goodpastpart))

    ewrongenpastpart, metalabel = makewrongenpastpart(stem, stemFS, takesge, prefix='e')
    triples.append((ewrongenpastpart, metalabel, goodpastpart))

    zerowrongenpastpart, metalabel = makewrongenpastpart(stem, stemFS, takesge, prefix='')
    triples.append((zerowrongenpastpart, metalabel, goodpastpart))

    if goodpastpart[-2:] == 'en':
        (ppstem, ppstemFS, takesge) = getstems(goodpastpart)
        wrongpastpart2, metalabel = makepastpart(ppstem, ppstemFS, takesge)
        wrongpastpart2a, metalabel = makewrongpastpart(ppstem, ppstemFS, takesge)

    else:
        wrongpastpart2, metalabel = goodpastpart, nolabel
        wrongpastpart2a, metalabel = goodpastpart, nolabel

    triples.append((wrongpastpart2, metalabel, goodpastpart))
    triples.append((wrongpastpart2a, metalabel, goodpastpart))

    egoodpastpart, metalabel = (goodpastpart[1:], nog) if goodpastpart[:2] == 'ge' else (goodpastpart, nolabel)
    triples.append((egoodpastpart, metalabel, goodpastpart))
    zerogoodpastpart, metalabel = (goodpastpart[2:], noge) if goodpastpart[:2] == 'ge' else (goodpastpart, nolabel)
    triples.append((zerogoodpastpart, metalabel, goodpastpart))

    # regular1 = (regularpastsg, regularpastpl, pastparticiple, pastpartwithe, wrongpastpart, wrongpastpart2, wrongpastpart2a, wrongenpastpart)
    # goodforms1 = (goodpastsg, goodpastpl, goodpastpart, goodpastpartwithe, goodpastpart, goodpastpart, goodpastpart, goodpastpart)
    # metanames1 = (overgen, overgen, overgen, overgen, wrongovergen, wrongovergen, wrongovergen, wrongovergen,)
    #
    # regular2 = (epastparticiple, zeropastparticiple, epastpartwithe, zeropastpartwithe, ewrongpastpart,
    #             zerowrongpastpart, ewrongenpastpart, zerowrongenpastpart, egoodpastpart, zerogoodpastpart)
    # goodforms2 = (goodpastpart, goodpastpart, goodpastpartwithe, goodpastpartwithe, goodpastpart,
    #               goodpastpart, goodpastpart, goodpastpart, goodpastpart, goodpastpart)
    #
    # regular = regular1 + regular2
    # goodforms = goodforms1 + goodforms2
    # return [regular, goodforms]
    return triples

# initialisation


# read the correctionfile
# eigenlijk nog de inflectiecode erbij@@@

#: The dictionary *correction* consists of items with a string as key and a tuple of
#: two strings (corrected form, metadata) as value. This dictionary is filled by
#: reading from the file with the name in the constant *correctionfilename* upon
#: initialisation of the module *deregularise*
correction: Dict[str, Tuple[str, str]] = {}
correctionfile = open(os.path.join(settings.SD_DIR, 'data', correctionfilename), 'r', encoding='utf8')
myreader = csv.reader(correctionfile, delimiter=tab)
for row in myreader:
    wrong = row[0]
    good = row[1]
    meta = row[2]
    correction[wrong] = good, meta
